import datetime

name = "CARP-S"
package_name = "carps"
author = "AutoML"
author_email = "smac@ai.uni-hannover.de"
description = "Benchmark SMAC"
url = "https://www.automl.org"
project_urls = {
    "Documentation": "https://AutoML.github.io/SMACBenchmarking/main",
    "Source Code": "https://github.com/AutoML/carps",
}
copyright = f"Copyright {datetime.date.today().strftime('%Y')}, AutoML"
version = "0.0.1"
